@@ -1,19 +1,20 @@
 from __future__ import (absolute_import, print_function, division,
                         unicode_literals)
 
+import _io
 import inspect
 import json as json_module
+import logging
 import re
-
-import _io
 import six
 
 from collections import namedtuple, Sequence, Sized
 from functools import update_wrapper
 from cookies import Cookies
-from requests.utils import cookiejar_from_dict
+from requests.adapters import HTTPAdapter
 from requests.exceptions import ConnectionError
 from requests.sessions import REDIRECT_STATI
+from requests.utils import cookiejar_from_dict
 
 try:
     from requests.packages.urllib3.response import HTTPResponse
@@ -22,8 +23,9 @@ except ImportError:
 
 if six.PY2:
     from urlparse import urlparse, parse_qsl, urlsplit, urlunsplit
+    from urllib import quote
 else:
-    from urllib.parse import urlparse, parse_qsl, urlsplit, urlunsplit
+    from urllib.parse import urlparse, parse_qsl, urlsplit, urlunsplit, quote
 
 if six.PY2:
     try:
@@ -33,21 +35,64 @@ if six.PY2:
 else:
     from io import BytesIO as BufferIO
 
+try:
+    from unittest import mock as std_mock
+except ImportError:
+    import mock as std_mock
+
+try:
+    Pattern = re._pattern_type
+except AttributeError:
+    # Python 3.7
+    Pattern = re.Pattern
+
 UNSET = object()
 
 Call = namedtuple('Call', ['request', 'response'])
 
+_real_send = HTTPAdapter.send
+
 _wrapper_template = """\
 def wrapper%(signature)s:
     with responses:
         return func%(funcargs)s
 """
 
+logger = logging.getLogger('responses')
+
 
 def _is_string(s):
     return isinstance(s, six.string_types)
 
 
+def _has_unicode(s):
+    return any(ord(char) > 128 for char in s)
+
+
+def _clean_unicode(url):
+    # Clean up domain names, which use punycode to handle unicode chars
+    urllist = list(urlsplit(url))
+    netloc = urllist[1]
+    if _has_unicode(netloc):
+        domains = netloc.split('.')
+        for i, d in enumerate(domains):
+            if _has_unicode(d):
+                d = 'xn--' + d.encode('punycode').decode('ascii')
+                domains[i] = d
+        urllist[1] = '.'.join(domains)
+        url = urlunsplit(urllist)
+
+    # Clean up path/query/params, which use url-encoding to handle unicode chars
+    if isinstance(url.encode('utf8'), six.string_types):
+        url = url.encode('utf8')
+    chars = list(url)
+    for i, x in enumerate(chars):
+        if ord(x) > 128:
+            chars[i] = quote(x)
+
+    return ''.join(chars)
+
+
 def _is_redirect(response):
     try:
         # 2.0.0 <= requests <= 2.2
@@ -63,7 +108,11 @@ def _is_redirect(response):
 def get_wrapped(func, wrapper_template, evaldict):
     # Preserve the argspec for the wrapped function so that testing
     # tools such as pytest can continue to use their fixture injection.
-    args, a, kw, defaults = inspect.getargspec(func)
+    if six.PY2:
+        args, a, kw, defaults = inspect.getargspec(func)
+    else:
+        args, a, kw, defaults, kwonlyargs, kwonlydefaults, annotations = \
+            inspect.getfullargspec(func)
 
     signature = inspect.formatargspec(args, a, kw, defaults)
     is_bound_method = hasattr(func, '__self__')
@@ -123,6 +172,8 @@ class BaseResponse(object):
     content_type = None
     headers = None
 
+    stream = False
+
     def __init__(self, method, url, match_querystring=False):
         self.method = method
         self.match_querystring = match_querystring
@@ -130,6 +181,26 @@ class BaseResponse(object):
         self.url = _ensure_url_default_path(url)
         self.call_count = 0
 
+    def __eq__(self, other):
+        if not isinstance(other, BaseResponse):
+            return False
+
+        if self.method != other.method:
+            return False
+
+        # Can't simply do a equality check on the objects directly here since __eq__ isn't
+        # implemented for regex. It might seem to work as regex is using a cache to return
+        # the same regex instances, but it doesn't in all cases.
+        self_url = self.url.pattern if isinstance(self.url,
+                                                  Pattern) else self.url
+        other_url = other.url.pattern if isinstance(other.url,
+                                                    Pattern) else other.url
+
+        return self_url == other_url
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
     def _url_matches_strict(self, url, other):
         url_parsed = urlparse(url)
         other_parsed = urlparse(other)
@@ -144,29 +215,25 @@ class BaseResponse(object):
             return False
 
         for (a_k, a_v), (b_k, b_v) in zip(url_qsl, other_qsl):
-            if not isinstance(a_k, six.text_type):
-                a_k = a_k.decode('utf-8')
-            if not isinstance(b_k, six.text_type):
-                b_k = b_k.decode('utf-8')
             if a_k != b_k:
                 return False
-            if not isinstance(a_v, six.text_type):
-                a_v = a_v.decode('utf-8')
-            if not isinstance(b_v, six.text_type):
-                b_v = b_v.decode('utf-8')
             if a_v != b_v:
                 return False
         return True
 
     def _url_matches(self, url, other, match_querystring=False):
         if _is_string(url):
+            if _has_unicode(url):
+                url = _clean_unicode(url)
+                if not isinstance(other, six.text_type):
+                    other = other.encode('ascii').decode('utf8')
             if match_querystring:
                 return self._url_matches_strict(url, other)
             else:
                 url_without_qs = url.split('?', 1)[0]
                 other_without_qs = other.split('?', 1)[0]
                 return url_without_qs == other_without_qs
-        elif isinstance(url, re._pattern_type) and url.match(other):
+        elif isinstance(url, Pattern) and url.match(other):
             return True
         else:
             return False
@@ -286,18 +353,23 @@ class RequestsMock(object):
 
     def __init__(self,
                  assert_all_requests_are_fired=True,
-                 response_callback=None):
+                 response_callback=None,
+                 passthru_prefixes=(),
+                 target='requests.adapters.HTTPAdapter.send'):
         self._calls = CallList()
         self.reset()
         self.assert_all_requests_are_fired = assert_all_requests_are_fired
         self.response_callback = response_callback
+        self.passthru_prefixes = tuple(passthru_prefixes)
+        self.target = target
 
     def reset(self):
         self._matches = []
         self._calls.reset()
 
-    def add(self,
-            method_or_response=None,
+    def add(
+            self,
+            method=None,  # method or ``Response``
             url=None,
             body='',
             adding_headers=None,
@@ -338,17 +410,68 @@ class RequestsMock(object):
         >>>     match_querystring=True
         >>> )
         """
-        if isinstance(method_or_response, BaseResponse):
-            self._matches.append(method_or_response)
+        if isinstance(method, BaseResponse):
+            self._matches.append(method)
             return
 
         if adding_headers is not None:
             kwargs.setdefault('headers', adding_headers)
 
-        if method_or_response:
-            kwargs['method'] = method_or_response
+        self._matches.append(
+            Response(method=method, url=url, body=body, **kwargs))
 
-        self._matches.append(Response(url=url, body=body, **kwargs))
+    def add_passthru(self, prefix):
+        """
+        Register a URL prefix to passthru any non-matching mock requests to.
+
+        For example, to allow any request to 'https://example.com', but require
+        mocks for the remainder, you would add the prefix as so:
+
+        >>> responses.add_passthru('https://example.com')
+        """
+        if _has_unicode(prefix):
+            prefix = _clean_unicode(prefix)
+        self.passthru_prefixes += (prefix, )
+
+    def remove(self, method_or_response=None, url=None):
+        """
+        Removes a response previously added using ``add()``, identified
+        either by a response object inheriting ``BaseResponse`` or
+        ``method`` and ``url``. Removes all matching responses.
+
+        >>> response.add(responses.GET, 'http://example.org')
+        >>> response.remove(responses.GET, 'http://example.org')
+        """
+        if isinstance(method_or_response, BaseResponse):
+            response = method_or_response
+        else:
+            response = BaseResponse(method=method_or_response, url=url)
+
+        while response in self._matches:
+            self._matches.remove(response)
+
+    def replace(self,
+                method_or_response=None,
+                url=None,
+                body='',
+                *args,
+                **kwargs):
+        """
+        Replaces a response previously added using ``add()``. The signature
+        is identical to ``add()``. The response is identified using ``method``
+        and ``url``, and the first matching response is replaced.
+
+        >>> responses.add(responses.GET, 'http://example.org', json={'data': 1})
+        >>> responses.replace(responses.GET, 'http://example.org', json={'data': 2})
+        """
+        if isinstance(method_or_response, BaseResponse):
+            response = method_or_response
+        else:
+            response = Response(
+                method=method_or_response, url=url, body=body, **kwargs)
+
+        index = self._matches.index(response)
+        self._matches[index] = response
 
     def add_callback(self,
                      method,
@@ -386,21 +509,30 @@ class RequestsMock(object):
         return get_wrapped(func, _wrapper_template, evaldict)
 
     def _find_match(self, request):
-        for match in self._matches:
+        found = None
+        found_match = None
+        for i, match in enumerate(self._matches):
             if match.matches(request):
-                break
-        else:
-            return None
-        # shift matches
-        self._matches = self._matches[1:]
-        self._matches.append(match)
-        return match
+                if found is None:
+                    found = i
+                    found_match = match
+                else:
+                    # Multiple matches found.  Remove & return the first match.
+                    return self._matches.pop(found)
+        return found_match
 
     def _on_request(self, adapter, request, **kwargs):
         match = self._find_match(request)
         resp_callback = self.response_callback
 
         if match is None:
+            if request.url.startswith(self.passthru_prefixes):
+                logger.info(
+                    'request.allowed-passthru', extra={
+                        'url': request.url,
+                    })
+                return _real_send(adapter, request)
+
             error_msg = 'Connection refused: {0} {1}'.format(
                 request.method, request.url)
             response = ConnectionError(error_msg)
@@ -436,16 +568,10 @@ class RequestsMock(object):
         return response
 
     def start(self):
-        try:
-            from unittest import mock
-        except ImportError:
-            import mock
-
         def unbound_on_send(adapter, request, *a, **kwargs):
             return self._on_request(adapter, request, *a, **kwargs)
 
-        self._patcher = mock.patch('requests.adapters.HTTPAdapter.send',
-                                   unbound_on_send)
+        self._patcher = std_mock.patch(target=self.target, new=unbound_on_send)
         self._patcher.start()
 
     def stop(self, allow_assert=True):
